import numpy as np
import matplotlib.pyplot as plt
from oct2py import Oct2Py
import os
import sys
import plotly.io
import plotly.graph_objects as go
from scipy.interpolate import LinearNDInterpolator
import math
from datetime import datetime
import copy

from .utils import rotate
from .prior import Prior
from .survey import GravitySurvey, GradientSurvey


class Fault:
    """
    Class that creates fault objects for gravity anaylsis. We assume that all input distacnes are in terms o m and the gravity is in mGal.
    Parameters
    ----------
        parameters: dict
            The parameters defining the fault.
        density: float
            The density contrast between the two layers that are separated by the faulted surface
        parameterised_model: np.array
            The parameterised model for the fault.
        parameter_labels: list
            If None, it defaults to the keys in the parameters dictionary
        displacement_model: np.array
            The model of the vertical displacement of the surface due to the fault.
        grid_separation: float
            The distance between each point on the grid that the fault is defined on
        grid_length: float
            The total are covered by the model is grid_length x grid_length
    """
    def __init__(self, parameters=None, density=None, parameterised_model=None, displacement_model=None, grid_separation=None, grid_length=None, parameter_labels=None):
        self.parameter_labels = parameter_labels
        self.fourier_domain_model = None
        self.grid = None
        self.parameters = parameters
        self.density = density
        self.displacement_model = displacement_model
        self.parameterised_model = parameterised_model
        self.grid_length = grid_length
        self.grid_separation = grid_separation

    def __setattr__(self, name, value):
        if name == 'parameters':
            if value is not None:
                default_keys = ["x1", "x2", "y1", "y2", "cx", "cy", "alpha", "l", "DL_ratio", "Extent_ratio", "dip", "depth"]
                for df in default_keys:
                    value.setdefault(df, None)
                if not isinstance(value, dict):
                    raise ValueError("parameters has to be a dictionary")
                if self.parameter_labels is None:
                    self.parameter_labels = list(value.keys())
                for i, key in enumerate(default_keys):
                    super().__setattr__(key, value[key])
        if name == 'displacement_model':
            if value is not None:
                if self.grid is None:
                    raise ValueError('Please provide the grid alongside the displacement model.')
                if np.shape(self.grid)[:2] != np.shape(value)[:2]:
                    raise ValueError('The number of grid locations is not the same as displacement values in displacement_model')
        super().__setattr__(name, value)

    def translate_to_grid(self, grid_separation=None, grid_length=None):
        """
        Translates the fault parameters to a displacement model.
        """
        octave = Oct2Py()
        octave.addpath(os.path.join(os.path.dirname(__file__), "faultmodel_m/"))
        self.translate_coords()

        if None in [self.parameters[key] for key in ['Extent_ratio', 'DL_ratio', 'dip']]:
            raise ValueError("Provide the Extent_ratio, the DL_ratio and the dip for the fault")
        if grid_separation is None:
            grid_separation = self.grid_separation
            if grid_separation is None:
                raise ValueError("Need to provide grid_separation.")
        if grid_length is None:
            grid_length = self.grid_length
            if grid_length is None:
                raise ValueError("Need to provide grid_length.")
        if position_at_centre:
            trace = np.array([[self.x1-self.cx, self.y1-self.cy],[self.x2-self.cx, self.y2-self.cy]])
            trace = trace+1e-15
        else:
            trace = np.array([[self.x1, self.y1], [self.x2, self.y2]])
        model, X, Y = octave.Model_FaultGrid(grid_length, grid_separation, trace, self.DL_ratio, self.Extent_ratio, self.dip, nout=3)
        octave.exit()
        self.grid = np.concatenate((np.expand_dims(X, axis=2), np.expand_dims(Y, axis=2)), axis=2)
        if np.isnan(model).any():
            print('found nan')
        self.displacement_model = model
        return model, X, Y

    def translate_to_parameterised_model(self, keys=None):
        """
        Translates the parameters to a parametersied model that can be used for training.
        """
        self.translate_coords()
        if keys is None:
            keys = self.parameter_labels
        self.parameterised_model = np.array([self.parameters[key] for key in keys])
        return self.parameterised_model

    def translate_to_parameters(self, defaults=None):
        """
        Takes the parameterised model and sets the corresponding elements in the parameters dictionary.
        Parameters
        ---------
            defaults: dict
                If it is set, then the parameters dictionary values that are None are set to these default values. This is so that not all parameters needs to be included in the parameterised_model (in case we don't want to infer them), but the fault can still be constructed.
        """
        keys = self.parameter_labels
        parameters_dict = dict.fromkeys(keys)
        for p, pl in enumerate(keys):
            parameters_dict[pl] = self.parameterised_model[p]
        self.parameters = parameters_dict
        if defaults is not None:
            for k in self.parameters.keys():
                if self.parameters[k] is None and any([k==dk for dk in defaults.keys()]):
                    self.parameters[k] = defaults[k]
                    self.__setattr__(k, defaults[k])


    def forward_model(self, num_components=50, depth=None, survey_coordinates=None, remove_min=True):
        """
        The fourier domain forward model. based on R.L Parker (1972)
        Takes the displacement model, depth and density contrast and turns it into a gravity signal on the surface. The surface is assume to be completely flat.
        Parameters
        ----------
            num_compontnents: int
                The number of fourier components to consider
            depth: float
                depth of the fault. If not give, try to use the depth from the parameter dictionary. The units are km.
            survey_coordinates: array
                [survey_points, 3] the survey coordiants in 3 dimensions, in the order of x, y z
                if not given, then the same grid is used as for the fault.
        """
        if self.displacement_model is None:
            self.translate_to_grid()
        if depth is None:
            depth = self.parameters["depth"]
            if depth is None:
                raise ValueError("Need to provide the depth of the fault")
        k_mag = self.make_k_vector()
        if np.isnan(self.displacement_model).any():
            print('Found NaN in displacement model')
        # eq.5 in R. L. Parker (1972)
        N = num_components
        R1 = np.zeros(np.shape(self.displacement_model))
        for n in range(N):
            f1 = np.fft.fft2((self.displacement_model*1000.0)**(n+1)) # changing to m
            r1 = np.complex128(k_mag**(n)/math.factorial(n+1)*f1) # r represent spatial domain, k represent k domain
            R1 = R1+r1
        self.fourier_domain_model = R1
        G = 6.67430*10**(-11) # Nm**2kg**(-2)
        f_g = -2*np.pi*G*np.exp((-k_mag)*depth*1000.0)*R1*self.density
        g = np.fft.ifft2(f_g)
        g_vec = g.ravel()
        g_orig = np.real(g_vec) * 1e5 # changing to mGal
        if survey_coordinates is None:
            if remove_min:
                return g_orig-np.min(g_orig), R1
            else:
                return g_orig, R1
        else:
            coords = np.reshape(self.grid, (np.shape(self.grid)[0]*np.shape(self.grid)[1], np.shape(self.grid)[2]))
            func = LinearNDInterpolator(coords, g_orig)
            g_new = func(survey_coordinates[:,0], survey_coordinates[:,1])
            if remove_min:
                return g_new-np.min(g_new), R1
            else:
                return g_new, R1

    def forward_model_gradient(self, depth=None, survey_coordinates=None, remove_min=True, baseline=1e-5):
        if depth is None:
            depth = self.parameters["depth"]
            if depth is None:
                raise ValueError("Need to provide the depth of the fault")

        l = baseline
        gz_1, _ = self.forward_model(depth=depth+l/2, survey_coordinates=survey_coordinates, remove_min=False)
        gz_2, _ = self.forward_model(depth=depth-l/2, survey_coordinates=survey_coordinates, remove_min=False)
        G_zz = (gz_1 - gz_2)/(l*1000) # mGal/m
        G_zz = G_zz * 1e4 # E
        if remove_min:
            return G_zz-np.min(G_zz)
        else:
            return G_zz

    def forward_model_from_fourier(self, depth=None, survey_coordinates=None):
        if self.fourier_domain_model is None:
            raise ValueError('fourier_model attribute not found')
        if depth is None:
            depth = self.parameters["depth"]
            if depth is None:
                raise ValueError("Need to provide the depth of the fault")
        else:
            self.depth = depth
            self.parameters['depth'] = depth
        k_mag = self.make_k_vector()
        depth = self.parameters['depth']
        R1 = self.fourier_domain_model
        G = 6.67430*10**(-11) # Nm**2kg**(-2)
        f_g = -2*np.pi*G*np.exp((-k_mag)*depth*1000.0)*R1*self.density
        g = np.fft.ifft2(f_g)
        g_vec = g.ravel()
        g_orig = np.real(g_vec) * 1e5 # changing to mGal
        if survey_coordinates is None:
            return g_orig, R1
        else:
            coords = np.reshape(self.grid, (np.shape(self.grid)[0]*np.shape(self.grid)[1], np.shape(self.grid)[2]))
            func = LinearNDInterpolator(coords, g_orig)
            g_new = func(survey_coordinates[:,0], survey_coordinates[:,1])
            return g_new, R1

    # ------------------ Plotting ---------------------------------
    def plot_3D_surface(self, survey_coordinates=None, filename='', depth=None):
        """
        Makes 3D plot of the fault.
        Parameters
        ----------
            filename : str
                The location where the image will be saved. The file format is infered from this string, can either be .png or .html
            depth: float
                The depth of the fault. If not given, try to get from parameters dictionary or set to 0.
        """
        plotly.io.templates.default = 'plotly_white'
        if self.displacement_model is None:
            self.translate_to_grid()
        X = self.grid[:,:,0]
        Y = self.grid[:,:,1] # changing to km
        model = self.displacement_model/1000.0
        if depth is None:
            depth = self.parameters["depth"]
            if depth is None:
                print("Plotting at 0 depth")
                depth = 0.0
        depth = depth
        model = np.reshape(model, np.shape(X))
        aspect_ratios = [1, 1, 0.2]
        fig = go.Figure(data=[go.Surface(z=model-depth, x=X, y=Y, cmax=np.max(-depth)-depth/100, cmin=np.min(-depth)+depth/100,
                                      contours = {"z": {"show": True, "start": np.min(model-depth), "end": np.max(model-depth), "size": (np.max(model-depth)-np.min(model-depth))/10, "color":"black"}},
                                      colorbar={"title": 'z [km]'})])

        if not survey_coordinates is None:
            fig.add_scatter3d(x=survey_coordinates[:,0], y=survey_coordinates[:,1], z=survey_coordinates[:,2], mode='markers', marker={'color': 'black', 'size': 1, 'opacity': 0.2}, showlegend=None)
            fig.update_layout(coloraxis_showscale=False)

        fig.update_scenes(aspectratio={"x":aspect_ratios[0], "y":aspect_ratios[1], "z":aspect_ratios[2]},
                          zaxis={"nticks": 2, "range": [np.min(model-depth), 0.01]},
                          xaxis={'range': [np.min(X), np.max(X)]},
                          yaxis={'range': [np.min(Y), np.max(Y)]},
                          xaxis_title="x [km]",
                          yaxis_title="y [km]",
                          zaxis_title="z [km]"
        )
        camera = dict(
            eye=dict(x=0.8, y=0.8, z=0.6))
        fig.update_layout(scene_camera=camera)
        #fig.update_layout({"paper_bgcolor": "rgba(0, 0, 0, 0)",
        #                    "plot_bgcolor": "rgba(0, 0, 0, 0)",})
        if filename[-5:] == '.html':
            fig.write_html(filename)
        elif filename[-4:] == '.png':
            fig.write_image(filename)
        else:
            raise ValueError("Only .html and .png file extensions are allowed")
        plt.close()

    def plot_pixels(self, filename='', survey_coordinates=None):
        """
        Creates a simple pixelised image of the survey. Can only be done for gridded data.
        """
        if self.displacement_model is None:
            raise ValueError("displacement model not given")


        plt.imshow(np.reshape(self.displacement_model, np.shape(self.grid[:,:,0])), extent=(np.min(self.grid[:,:,0]), np.max(self.grid[:,:,0]), np.max(self.grid[:,:,1]), np.min(self.grid[:,:,1])))
        plt.colorbar(label='km')
        if survey_coordinates is not None:
            plt.scatter(survey_coordinates[:,0], survey_coordinates[:,1], s=1, marker='o', color='black')
        plt.scatter(self.cx, self.cy, color='red', marker='x', s=2)
        plt.xlabel('x [km]')
        plt.ylabel('y [km]')
        plt.savefig(filename)
        plt.close()

    # ------------------ Utilities ---------------------------------
    def translate_coords(self):
        """
        Defines translation between end points x1, y1, x2, y2 and central location, lenght and rotation angle (cx, cy, l, alpha)
        """
        keys = ["cx", "cy", "x1", "y1", "x2", "y2", "l", "alpha"]
        if not None in [self.parameters[key] for key in keys]:
            return
        elif None in [self.parameters[key] for key in ["x1", "y1", "x2", "y2"]]:
            if None in [self.parameters[key] for key in ["cx", "cy", "l", "alpha"]]:
                raise ValueError("Need to provide one set of location parameters. Either (cx, cy, l, alpha) or (x1, y1, x2, y2)")
            self.x1 = self.l/2*np.cos(self.alpha)+self.cx
            self.y1 = self.l/2*np.sin(self.alpha)+self.cy
            self.x2 = -self.l/2*np.cos(self.alpha)+self.cx
            self.y2 = -self.l/2*np.sin(self.alpha)+self.cy
            self.parameters['x1'] = self.x1
            self.parameters['x2'] = self.x2
            self.parameters['y1'] = self.y1
            self.parameters['y2'] = self.y2
        elif None in [self.parameters[key] for key in ["cx", "cy", "l", "alpha"]]:
            if None in [self.parameters[key] for key in ["x1", "y1", "x2", "y2"]]:
                raise ValueError("Need to provide one set of location parameters. Either (cx, cy, l, alpha) or (x1, y1, x2, y2)")
            self.l = np.sqrt((self.x2-self.x1)**2+(self.y2-self.y1)**2)
            self.cx = (self.x1+self.x2)/2
            self.cy = (self.y1+self.y2)/2
            self.alpha = np.arccos(2*(self.x1-self.cx)/self.l)
            if self.cy > self.y1:
                self.alpha = 2*np.pi-self.alpha
            self.parameters['alpha'] = self.alpha
            self.parameters['l'] = self.l
            self.parameters['cx'] = self.cx
            self.parameters['cy'] = self.cy
        return

    def make_k_vector(self):
        X = self.grid[:,:,0]*1000.0 # changing to m
        Y = self.grid[:,:,1]*1000.0
        numrows = np.shape(X)[0]
        numcolumns = np.shape(Y)[0]
        longx = np.max(X)-np.min(X)
        longy = np.max(Y)-np.min(Y)
        frequency = np.zeros((abs((numrows // 2) + 1), abs((numcolumns // 2) + 1)))

        for f in range(1, int((numrows/2) + 2)):
            for g in range(1, int((numcolumns/2) + 2)):
                frequency[f-1, g-1] = np.sqrt(((f-1) / longx) ** 2 + ((g-1) / longy) ** 2)

        frequency2 = np.fliplr(frequency)
        frequency3 = np.flipud(frequency)
        frequency4 = np.fliplr(np.flipud(frequency))

        entero = round(numcolumns / 2)
        if ((numcolumns / 2) - entero) == 0:
            frequency2 = np.delete(frequency2, 0, axis=1)
            frequency3 = np.delete(frequency3, 0, axis=0)
            frequency4 = np.delete(np.delete(frequency4, 0, axis=1), 0, axis=0)
            frequencytotal = np.concatenate((np.concatenate((frequency, frequency2), axis=1),
                                            np.concatenate((frequency3, frequency4), axis=1)))
        else:
            frequencytotal = np.concatenate((np.concatenate((frequency, frequency2), axis=1),
                                            np.concatenate((frequency3, frequency4), axis=1)))
        frequencytotal = frequencytotal[:-1, :]
        frequencytotal = frequencytotal[:, :-1]
        frequencytotal = frequencytotal * (2 * np.pi)
        return frequencytotal

    def change_grid_reference_frame(self, coordinate_centres=[0,0]):
        """
        Chenges the frame of reference of the grid coordiante system.
        Parameters
        ---------
            coordinate_centres: list of floats
                This should be in the current frame,
                Defines the centre point of the new coordinate system
        """
        x, y = coordinate_centres
        self.grid[:,:,0] = self.grid[:,:,0] - x
        self.grid[:,:,1] = self.grid[:,:,1] - y
        return

    def change_reference_frame(self, coordinate_centres=[0,0]):
        """
        Chenges the frame of reference of the coordiante system.
        If the parameterised_model was made before, it is now reset to None and it needs to be made again.
        Parameters
        ---------
            coordinate_centres: list of floats
                This should be in the current frame,
                Defines the centre point of the new coordinate system
        """
        x, y = coordinate_centres
        self.parameters['cx'] = self.parameters['cx']-x
        self.parameters['cy'] = self.parameters['cy']-y
        self.cx = self.cx-x
        self.cy = self.cy-y
        self.parameterised_model = None
        self.grid[:,:,0] = self.grid[:,:,0] - x
        self.grid[:,:,1] = self.grid[:,:,1] - y
        return

class FaultDataset():
    """
    Class for making a data set of faults and corresponding gravity surveys.
    Attributes
    ----------
        size: int
            Number of data points
        priors: Prior
            class that defines the priors for the fault parameters
        model_framework: dict
            keys:   type: (str) currently not used for anything
                    parameters_to_include: (list) the parameter keys that we want to include in a training dataset
                    density: (float) the density contrast between the layers separated by the faulted surface
                    grid_length: (float) the total area covered by the dispalcement model of the fault
                    grid_separation: (float) the distance between each grid point
        survey_framework: dict
            keys:   noise_scale: (float) the standard deviation of the gaussian noise added to the survey
                    sizes: (list) length 3, the size of the survey area in the three different dimensions
                    survey_shape: (list) [num_x, num_y] or [num_x, num_y, num_z] the number of survey points in eac dimension
                    fault_parameters_as_centre: (bool) if True, the survey centre is location at the fault location parameters cx cy. The effect of this is that these parameters are then defined with respect to the survey position
        faults: list
            the list of the generated fault objects
        surveys: list
             the list of the generated surveys
    """
    def __init__(self, size: int, priors: Prior, model_framework={}, survey_framework={}):
        self.size = size
        self.priors = priors
        self.model_framework = model_framework
        self.survey_framework = survey_framework
        self.parameter_labels = priors.keys
        self.faults = None
        self.surveys = None

    def __setattr__(self, name, value):
        if name == 'model_framework':
            if not isinstance(value, dict):
                raise ValueError("Expected dict for model_framework.")
            value.setdefault("type", 'parameterised')
            value.setdefault("parameters_to_include", ['cx', 'cy', 'l', 'alpha'])
            value.setdefault("density", 800.0)
            value.setdefault("grid_length", None)
            value.setdefault("grid_separation", None)
            value.setdefault("reference_frame", "fault_grid")
        if name == 'survey_framework':
            if not isinstance(value, dict):
                raise ValueError("Expected dict for survey_framework.")
            value.setdefault("type", 'gravity')
            value.setdefault("noise_scale", 0.0)
            value.setdefault("sizes", [1,1,0])
            value.setdefault("survey_shape", [5,5])
            value.setdefault("fault_parameters_as_centre", False)
            if not isinstance(value["survey_shape"], list):
                raise ValueError("The survey shape has to be a list. Can have a single element")
        super().__setattr__(name, value)

    def make_dataset(self, parameters_dict=None, seed=None, location_augment_factor=1, slim_output=False):
        """
        Makes a data set of faults.
        Parameters
        ----------
            parameters_dict: dict
                If None, then the fault parameters are randomly generated from the prior.
                Otherwise the parameters are taken from the dictionary. (Default: None)
            parameter_keys_to_include: list
                The keys of the parameters to be incldued from the parameters_dict in the parametersied model.
        Output
        ------
            faults, surveys : two lists with the size of the data set.
        """
        total_time = datetime.now()
        if parameters_dict is None:
            parameters_dict = self.priors.sample(size=self.size, returntype='dict', seed=seed) # if the parameters dictionary is not passed to the function, then the prior is sampled
            dataset_size = self.size
        else:
            if self.priors.keys != list(parameters_dict.keys()):
                print("FYI: The prior and the parameter keys do not match")
            dataset_size = np.shape(parameters_dict[list(parameters_dict.keys())[0]])[0]

        # check that self.size is divisible by location_augment_factor:
        if self.size % location_augment_factor == 0:
            num_faults = int(self.size/location_augment_factor)
        else:
            raise ValueError('The size of the dataset has to be divisible by the location augment factor')

        # Making a set of faults first with only parameters
        faults = []
        start_time = datetime.now()
        for i in range(num_faults):
            fault_parameters = dict.fromkeys(self.priors.keys)
            for key in list(self.priors.keys):
                print(parameters_dict[key])
                fault_parameters[key] = parameters_dict[key][i]
            fault = Fault(parameters=fault_parameters, density=self.model_framework['density']) # the fault parameters are assigned to the class. This contains values directily sampled from the piror.
            fault.translate_coords()
            faults.append(fault)
            if i % 1000 == 0:
                print(f"{i}/{num_faults} Faults generated \t Time taken: {datetime.now()-start_time}")

        # Translating the faults into displacement models
        faults = self.translate_to_grid_vectorised(faults, grid_separation=self.model_framework['grid_separation'], grid_length=self.model_framework['grid_length'], position_at_centre=True) # we are always putting the fault in the middle, despite the cx and cy positions

        # Get the forward model
        fms = []
        fms_func = []
        for fault in faults:
            if self.survey_framework['type'] == 'gravity':
                g, _ = fault.forward_model()
            if self.survey_framework['type'] == 'gradient':
                g = fault.forward_model_gradient()
            fms.append(g)
            coords = np.reshape(fault.grid, (np.shape(fault.grid)[0]*np.shape(fault.grid)[1], np.shape(fault.grid)[2]))
            fms_func.append(LinearNDInterpolator(coords, g)) # doing interpolation here, as it is faster to just do once
            if slim_output:
                fault.grid = None
                fault.displacement_model=None

        # Making the surveys
        surveys = []
        start_time = datetime.now()
        for idx in range(self.size):

            # here we are allowing variable sizes of surveys. If a range of possible values are given, then we sample this range uniformly.
            survey_sizes = []
            for i, ss in enumerate(self.survey_framework['sizes']):
                if isinstance(ss, list):
                    np.random.seed(seed=seed)
                    survey_sizes.append(np.random.uniform(low=ss[0], high=ss[1], size=1)[0]) # sampling randomly from given survey range
                else:
                    survey_sizes.append(ss)
            # translating the survey sizes to ranges that the Survey class can understand
            survey_ranges = []
            for ss in survey_sizes:
                if ss != 0:
                    survey_ranges.append([-ss/2, ss/2])
                else:
                    survey_ranges.append([0])

            if self.survey_framework['fault_parameters_as_centre']: # we take the cx and cy parameters from the fault and place the survey centre there.
                survey_ranges[0] = survey_ranges[0]-parameters_dict['cx'][idx]
                survey_ranges[1] = survey_ranges[1]-parameters_dict['cy'][idx]

            # Construction the survey
            if self.survey_framework['type'] == 'gravity':
                survey = GravitySurvey(ranges=survey_ranges, survey_shape=self.survey_framework['survey_shape'])
            elif self.survey_framework['type'] == 'gradient':
                survey = GradientSurvey(ranges=survey_ranges, survey_shape=self.survey_framework['survey_shape'])
            else:
                raise ValueError('survey_framework type can only be gravity or gradient.')
            survey.make_survey()

            # Checking that the survey never lies outside the modelled fault area
            if survey.survey_coordinates[:,:2].max() > self.model_framework['grid_length']/2 or survey.survey_coordinates[:,:2].min() < -self.model_framework['grid_length']/2:
                print("Survey lies outside of grid limits.")
                continue

            # Making some noise
            noise_prior = Prior(distributions={"noise_scale" :self.survey_framework['noise_scale']})
            noise_scale = noise_prior.sample(size=1, returntype='dict')['noise_scale'][0]
            survey.noise_scale = noise_scale
            survey.make_noise()

            surveys.append(survey) # adding the complete survey to the list
            if idx % 1000 == 0:
                print(f"{idx}/{self.size} surveys generated \t Time taken: {datetime.now()-start_time}")

        # interpolating the forward modelled gravity onto the survey grids
        idx = 0
        for s in surveys:
            if idx == num_faults:
                idx = 0
            s.interpolate_data_onto_survey_coordinates(func=fms_func[idx], remove_min=True)
            idx = idx+1
        faults_new = []
        idx = 0
        for i in range(self.size):
            if idx == num_faults:
                idx = 0
            if slim_output:
                f = Fault(parameters = faults[idx].parameters)
            else:
                f = copy.deepcopy(faults[idx]) # the fault parameters are assigned to the class. This contains values directily sampled from the piror
            f.parameters['cx'] = parameters_dict['cx'][i]
            f.parameters['cy'] = parameters_dict['cy'][i]
            f.cx = parameters_dict['cx'][i]
            f.cy = parameters_dict['cy'][i]
            faults_new.append(f)
            idx = idx+1
            # a number of the faults are gonna have the same length, alpha and depth, but the centre location will always be different
        faults = None
        # changing the reference frame of the coordinates to the survey, if requested
        # shifting both the survey coordinates and the fault grid by the same amount, so that the centre point of the survey is 0, but their relative position remains the same
        if self.model_framework['reference_frame'] == 'survey':
            for i in range(self.size):
                if not slim_output:
                    faults_new[i].change_grid_reference_frame(coordinate_centres=[-faults_new[i].cx, -faults_new[i].cy])
                surveys[i].change_reference_frame(coordinate_centres=[-faults_new[i].cx, -faults_new[i].cy])

        # If we are making a parameterised data set then the faults are translated into parameterised models
        if self.model_framework['type'] == 'parameterised':
            for i in range(self.size):
                faults_new[i].translate_to_parameterised_model(keys=self.model_framework['parameters_to_include'])

        self.surveys = surveys
        self.faults = faults_new
        print(f"Dataset made. Total time taken: {datetime.now()-total_time}")
        if slim_output:
            return_params = dict.fromkeys(self.priors.keys)
            for k in self.priors.keys:
                return_params[k] = np.array([f.parameters[k] for f in self.faults])
            return {'gravity': np.array([s.gravity for s in self.surveys]), 'survey_coordinates': np.array([s.survey_coordinates for s in self.surveys]), 'survey_ranges': [s.ranges for s in self.surveys], 'noise': [s.noise for s in self.surveys]} | return_params
        else:
            return self.faults, self.surveys
#
    def augment_depth(self):
        print(f"Making {self.size} depth augmented data..")
        start_time = datetime.now()
        parameters_dict = self.priors.sample(size=self.size, returntype='dict')
        depths = parameters_dict['depth']
        for i in range(self.size):
            survey = self.surveys[i]
            self.surveys[i].gravity, _ = self.faults[i].forward_model_from_fourier(depth=depths[i], survey_coordinates = self.surveys[i].survey_coordinates.copy())
            self.surveys[i].make_noise() # changing the noise realisation too
            if i % 1000 == 0:
                print(f"{i}/{self.size} data augmented \t Time taken: {datetime.now()-start_time}")

    def augment_depth_v2(self):
        print(f"Making {self.size} depth augmented data..")
        start_time = datetime.now()
        parameters_dict = self.priors.sample(size=self.size, returntype='dict')
        depths = parameters_dict['depth']
        for i in range(self.size):
            fault_coords_initial = copy.deepcopy(self.faults[i].grid)
            survey_coords_initial = copy.deepcopy(self.surveys[i].survey_coordinates)
            self.faults[i].change_grid_reference_frame(coordinate_centres=[self.faults[i].cx, self.faults[i].cy])
            self.surveys[i].change_reference_frame(coordinate_centres=[self.faults[i].cx, self.faults[i].cy])
            fault_coords_intermittent = copy.deepcopy(self.faults[i].grid)
            survey_coords_intermittent= copy.deepcopy(self.surveys[i].survey_coordinates)
            print('intermittent', self.faults[i].cx, self.faults[i].cy)
            print(fault_coords_intermittent-fault_coords_initial)
            print(survey_coords_intermittent-survey_coords_initial)
            self.surveys[i].gravity, _ = self.faults[i].forward_model_from_fourier(depth=depths[i], survey_coordinates = self.surveys[i].survey_coordinates.copy())
            self.surveys[i].make_noise() # changing the noise realisation too
            self.faults[i].change_grid_reference_frame(coordinate_centres=[-self.faults[i].cx, -self.faults[i].cy])
            self.surveys[i].change_reference_frame(coordinate_centres=[-self.faults[i].cx, -self.faults[i].cy])
            fault_coords_final = copy.deepcopy(self.faults[i].grid)
            survey_coords_final = copy.deepcopy(self.surveys[i].survey_coordinates)
            print('final')
            print(fault_coords_final-fault_coords_initial)
            print(survey_coords_final-survey_coords_initial)
            if i % 1000 == 0:
                print(f"{i}/{self.size} data augmented \t Time taken: {datetime.now()-start_time}")

    def translate_to_grid_vectorised(self, faults, grid_separation=None, grid_length=None, position_at_centre=False):
        """
        Translates fault parameters to displacement models.
        Same as for the Fault class, but looped. This prevents octave from needing to be closed and reopened all the time.
        """
        oc = Oct2Py()
        oc.addpath(os.path.join(os.path.dirname(__file__), "faultmodel_m/"))
        for i in range(len(faults)):
            fault = faults[i]

            if None in [fault.parameters[key] for key in ['Extent_ratio', 'DL_ratio', 'dip']]:
                raise ValueError("Provide the Extent_ratio, the DL_ratio and the dip for the fault")
            if grid_separation is None:
                grid_separation = fault.grid_separation
                if grid_separation is None:
                    raise ValueError("Need to provide grid_separation.")
            if grid_length is None:
                grid_length = fault.grid_length
                if grid_length is None:
                    raise ValueError("Need to provide grid_length.")
            if position_at_centre:
                trace = np.array([[fault.x1-fault.cx, fault.y1-fault.cy],[fault.x2-fault.cx, fault.y2-fault.cy]])
                trace = trace+1e-15
            else:
                trace = np.array([[fault.x1, fault.y1],[fault.x2, fault.y2]])
            model, X, Y = oc.Model_FaultGrid(grid_length, grid_separation, trace, fault.DL_ratio, fault.Extent_ratio, fault.dip, nout=3)
            fault.grid = np.concatenate((np.expand_dims(X, axis=2), np.expand_dims(Y, axis=2)), axis=2)
            fault.displacement_model = model
            if i % 1000 == 0:
                print(f"{i}/{self.size} Faults translated to displacement model")
        oc.exit()
        return faults


    def make_data_for_network(self, survey_coordinates_to_include=[], model_parameters_to_include=['cx', 'cy', 'l', 'alpha'], add_noise=True, mix_survey_order=False):
        """
        Parameters
        ----------
            survey_coordinates_to_include: list
                elements can be 'x', 'y', 'z': the array of measurement coordinates are included for each data point
                            'x_range' 'y_range' 'z_range': only the range of the coordinates are included. Lower and upper limits
                            'noise_scale': the scale of the gaussian noise is included (Default: [])
            mix_survey_order: bool
                if True, the order of the survey measurement points are shuffled randomly. (Default: False)
            add_noise: bool
                if True, measurement noise is added to the surveys with the define noise_scale. (Default: True)
            model_info_to_include: list
                elements can be 'noise_scale': the background noise scale, used for voxelised representation. (Default: [])
        Output
        ------
            data: list
                List of arrays with information to include in the desied data set. The first element of this list is always the model of the box (either parameterised or voxelised)
            condition: list
                List of arrays with information to include int he desired data set. The first element of this list is always the gravity array.
        """
        for f in self.faults:
            f.translate_to_parameterised_model(keys = model_parameters_to_include)

        data = []
        for key in model_parameters_to_include:
            arr = np.array([self.faults[i].parameters[key] for i in range(self.size)])[...,np.newaxis]
            data.append(arr)
        print(np.shape(data))

#        if any([l=='noise_scale' for l in model_info_to_include]):
#            data.append(np.expand_dims(np.array([self.faults[i].background_noise_scale for i in range(self.size)]), axis=1))
#
        conditional_gz = np.array([self.surveys[i].gravity for i in range(self.size)])
        conditional_coordinates = np.array([self.surveys[i].survey_coordinates for i in range(self.size)])

        if add_noise:
            noise = np.array([self.surveys[i].noise for i in range(self.size)])
            conditional_gz = conditional_gz+noise

        if mix_survey_order:
            i_arr = np.arange(np.shape(conditional_gz)[1])
            for i in range(self.size):
                np.random.shuffle(i_arr)
                conditional_gz[i,:] = conditional_gz[i,:][i_arr]
                for x in range(np.shape(conditional_coordinates)[2]):
                    conditional_coordinates[i,:,x] = conditional_coordinates[i,:,x][i_arr]

        conditional = [conditional_gz]

        labels = ['x', 'y', 'z']
        for idx, label in enumerate(labels):
            if any([l==label for l in survey_coordinates_to_include]):
                conditional.append(conditional_coordinates[:,:,idx])
        labels = ['x_ranges', 'y_ranges', 'z_ranges']
        for idx, label in enumerate(labels):
            if any([l==label for l in survey_coordinates_to_include]):
                conditional.append(np.array([self.surveys[i].ranges[idx] for i in range(self.size)]))
        if any([c=='survey_width_ratio' for c in survey_coordinates_to_include]):
            conditional.append(np.expand_dims(np.array([((self.surveys[i].ranges[0][1]-self.surveys[i].ranges[0][0])/(self.surveys[i].ranges[1][1]-self.surveys[i].ranges[1][0])) for i in range(self.size)]), axis=1))
        if any([l=='noise_scale' for l in survey_coordinates_to_include]):
            conditional.append(np.expand_dims(np.array([self.surveys[i].noise_scale for i in range(self.size)]), axis=1))

        return data, conditional
   
